/** 
 * \file halfedge_attribute.cpp
 *  
 * \brief  Implementation   of  the  class   HalfedgeAttribute,  which
 * represents a set of  halfedge attributes of the underlying triangle
 * surface mesh of a  Parametric Pseudo-Surface (PPS) constructed from
 * a Loop subdivision surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "halfedge_attribute.h"  // HalfedgeAttribute

#include <cassert>               // assert


/** 
* \defgroup PPSFROMLOOPNameSpace Namespace ppsfromloop.
* @{
*/

/**
* \namespace ppsfromloop
*
* \brief  The  namespace   ppsfromloop  contains  the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS) that approximates  the geometry of a given Loop
* subdivision surface.
*/

namespace ppsfromloop {


  /**
   * \fn unsigned HalfedgeAttribute::compute_pps_id() const
   *
   * \brief  Computes and  returns  the PPS  identifier  (ID) of  this
   * half-edge.
   *
   * \return The PPS identifier of this half-edge.
   */
  unsigned
  HalfedgeAttribute::compute_pps_id() const
  {
    //
    // This attribute must be owned by some half-edge.
    //
    assert( get_owner() != 0 ) ;

    unsigned i = 0 ;
    Halfedge* h = get_owner()->get_origin()->get_halfedge() ;

    while ( h != get_owner() ) {
      i++ ;
      h = h->get_prev()->get_mate() ;
    }

    return i ;
  }


  /**
   * \fn unsigned HalfedgeAttribute::compute_origin_vertex_degree() const
   *
   * \brief Computes  and returns the  degree of the origin  vertex of
   * this half-edge.
   *
   * \return The degree of the origin vertex of this half-edge.
   */
  unsigned
  HalfedgeAttribute::compute_origin_vertex_degree() const
  {
    //
    // This attribute must be owned by some half-edge.
    //
    assert( get_owner() != 0 ) ;

    unsigned i = 0 ;
    Halfedge* h = get_owner() ;
    do {
      ++i ;

      h = h->get_prev()->get_mate() ;
    }
    while ( h != get_owner() ) ;

    return i ;
  }

}

/** @} */ //end of group class.
