/** 
 * \file ppsfromloop.cpp
 *  
 * \brief  Definition  of  the  class PPSfromLOOP  that  represents  a
 * Parametric  Pseudo Surface, which  approximates a  Loop subdivision
 * surface  defined  on  a  triangle  mesh  represented  by  a  Doubly
 * Connected List (DCEL).
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "ppsfromloop.h"  // PPSfromLOOP

#include <cassert>        // assert
#include <cmath>          // fabs


/** 
* \defgroup PPSFROMLOOPNameSpace Namespace ppsfromloop.
* @{
*/

/**
* \namespace ppsfromloop
*
* \brief  The  namespace   ppsfromloop  contains  the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface  (PPS)  that approximates  the  geometry  of a  given
* Loop's subdivision surface.
*/

namespace ppsfromloop {

  using pps::PPS ;

  /**
   * \fn PPSfromLOOP::PPSfromLOOP( Surface* mesh , std::string fn )
   *
   * \brief Creates an instance of this PPSfromLOOP class. 
   *
   * \param mesh A Pointer to a surface mesh represented by a DCEL.
   * \param fn Name of the file containing the evaluator data table.
   *
   */
  PPSfromLOOP::PPSfromLOOP(
			   Surface* mesh ,
			   std::string fn 
			  )
    : PPS< Surface >( mesh )
  {
    //
    // Create the Loop surface evaluator.
    //

    //
    // Load in the evaluator data table.
    //
    _evstruct = new EvalStruct( fn.c_str() ) ;

    //
    // Sets the owner of each half-edge attribute set.
    //
    for ( EdgeIterator eit = edges_begin() ; !is_done( eit ) ; 
	  move_forward( eit ) ) {
      Edge* edge = get_edge( eit ) ;

      Halfedge* h1 = edge->get_first_halfedge() ;
      Halfedge* h2 = edge->get_second_halfedge() ;

      h1->get_attributes().set_owner( h1 ) ;
      h2->get_attributes().set_owner( h2 ) ; 
    }

    //
    // Create the patches associated with the mesh base faces.
    //
    build_surface_evaluator() ;

    return ;
  }


  /**
   * \fn ~PPSfromLOOP::PPSfromLOOP()
   *
   * \brief Destroys an instance of this class.
   */
  PPSfromLOOP::~PPSfromLOOP()
  {
    if ( _evstruct != 0 ) {
      delete _evstruct ;
    }

    return ;
  }


  /**
   * \fn void PPSfromLOOP::eval_surface( Face* face , double u , double v , double w , double& x , double& y , double& z ) const
   *
   * \brief  Computes  a  point   on  the  subdivision  surface  patch
   * associated with a given face of the PPS underlying triangle mesh.
   *
   * \param face Pointer to one face of the underlying mesh.
   * \param u First barycentric coordinate of a point on the face.
   * \param v Second barycentric coordinate of a point on the face.
   * \param w Third barycentric coordinate of a point on the face.
   * \param x First Cartesian coordinate of a point on the PPS image.
   * \param y Second Cartesian coordinate of a point on the PPS image.
   * \param z Third Cartesian coordinate of a point on the PPS image.
   */
  void
  PPSfromLOOP::eval_surface(
			    Face* face , 
			    double u ,
			    double v ,
			    double w ,
			    double& x ,
			    double& y ,
			    double& z
			   )
    const 
  {
    //
    // Make  sure  the  face  has  a Loop  subdivision  surface  patch
    // associated with it.
    //
    LOOPatch* patch = face->get_attributes().get_patch() ; 

    assert( patch != 0 ) ;

    //
    // Make sure the coordinates are non-negative and add up to 1.
    //
    assert( 
	   ( u >= 0 ) && ( u <= 1 ) && ( v >= 0 ) && 
	   ( v <= 1 ) && ( w >= 0 ) && ( w <= 1 ) &&
	   ( fabs( 1 - ( u + v + w ) ) <= 1e-15 )
	  ) ;
    //
    // Evaluate the patch.
    //
    patch->point( u , v , x , y , z ) ;

    return ;
  }


  /**
   * \fn void PPSfromLOOP::build_surface_evaluator()
   *
   * \brief Creates a Loop subdivision surface evaluator.
   */
  void
  PPSfromLOOP::build_surface_evaluator()
  {
    /**
     * For each face  of the given triangle surface  mesh, compute the
     * surface patch associated with  it in the given Loop subdivision
     * surface.
     */
    for ( FaceIterator fit = faces_begin() ; !is_done( fit ) ; move_forward( fit ) ) {

      // -------------------------------------------------------------
      // Find the extraordinary vertex of the current face, if any.
      //
      // Recall that there can be AT MOST ONE extraordinary vertex per
      // face.
      // -------------------------------------------------------------

      Face* face = get_face( fit ) ;

      Halfedge* he1 = face->get_halfedge() ;
      Halfedge* he2 = he1->get_next() ;
      Halfedge* he3 = he1->get_prev() ;

      unsigned nv1 = he1->get_attributes().get_origin_vertex_degree() ;
      unsigned nv2 = he2->get_attributes().get_origin_vertex_degree() ;
      unsigned nv3 = he3->get_attributes().get_origin_vertex_degree() ;

      Halfedge* he;
      unsigned nv;
      LOOPatch::BCORDER bo;

      assert( 
	     ( ( nv2 == 6 ) && ( nv3 == 6 ) ) ||
	     ( ( nv1 == 6 ) && ( nv3 == 6 ) ) ||
	     ( ( nv1 == 6 ) && ( nv2 == 6 ) )
	    ) ;

      if ( ( nv2 == 6 ) && ( nv3 == 6 ) ) {
        nv = nv1 ;
        he = he1 ;
        bo = LOOPatch::UVW ;
      }
      else if ( ( nv1 == 6 ) && ( nv3 == 6 ) ) {
        nv = nv2 ;
        he = he2 ;
        bo = LOOPatch::VWU ;
      }
      else {  // ( ( nv1 == 6 ) && ( nv2 == 6 ) )
        nv = nv3 ;
        he = he3 ;
        bo = LOOPatch::WUV ;
      }

      const unsigned int MAX_DEGREE = 
	unsigned( _evstruct->get_max_degree() ) ;

      assert( nv <= MAX_DEGREE ) ;

      // Compute  the "projected  points" of  the  subdivision surface
      // patch  associated with  the current  face of  the  given base
      // mesh.

      const unsigned int K = nv + 6;

      Halfedge** vhe = new Halfedge*[ K ] ;

      vhe[ 0 ] = he ;
      vhe[ 1 ] = he->get_next() ;
      he = he->get_mate()->get_next()->get_mate() ;

      for ( unsigned int i = 2 ; i <= nv ; i++ ) {
        vhe[ i ] = he ;
	he = he->get_next()->get_mate() ;
      }

      he = vhe[ 1 ]->get_mate()->get_next()->get_mate() ;

      for ( unsigned int i = nv + 1 ; i <= nv + 3 ; i++ ) {
        vhe[ i ] = he ;
        he = he->get_next()->get_mate() ;
      }

      he = vhe[ 1 ]->get_mate()->get_prev()->get_mate()->get_prev() ;

      for ( unsigned int i = nv + 4 ; i <= nv + 5 ; i++ ) {
        vhe[ i ] = he ;
	he = he->get_mate()->get_prev() ;
      }

      double* control_pts = new double[ K * 3 ] ;

      for ( unsigned int i = 0 ; i < K ; i++ ) {
        control_pts[ LOOPatch::index( 0 , i , K) ] = 
	  vhe[ i ]->get_origin()->x() ;
        control_pts[ LOOPatch::index( 1 , i , K) ] = 
	  vhe[ i ]->get_origin()->y() ;
        control_pts[ LOOPatch::index( 2 , i , K) ] = 
	  vhe[ i ]->get_origin()->z() ;
      }

      delete[] vhe ;

      double* proj_pts = _evstruct->proj_control_points( control_pts ,
							 int( nv ) ) ;

      delete[] control_pts ;

      //
      // Assign the  Loop subdivision  surface patch with  the current
      // face.
      //
      face->get_attributes().set_patch( 
		                       new LOOPatch(
				                    _evstruct ,
				                    nv ,
				                    bo ,
				                    proj_pts
				                   )
		                      ) ;

      delete[] proj_pts ;
    }

    return ;
  }

}

/** @} */ //end of group class.
