/** 
 * \file sampler.cpp
 *  
 * \brief Samples a PPS built from a Loop subdivision surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>              // std::cout, std::endl
#include <string>                // std::string
#include <cmath>                 // sqrt
#include <cstdlib>               // atoi

#include "surface.h"             // dcel::Surface

#include "pps.h"                 // pps::PPS
#include "ppssampler.h"          // pps::PPSsampler
#include "ppsfromloop.h"         // ppsfromloop::PPSfromLOOP

#include "reader.h"              // offlib::Reader
#include "writer.h"              // offlib::Writer

#include "vertex_attribute.h"    // ppsfromloop::VertexAttribute
#include "halfedge_attribute.h"  // ppsfromloop::HalfedgeAttribute
#include "face_attribute.h"      // ppsfromloop::FaceAttribute


/**
 * \fn int main( int argc , char *argv[ ] )
 *
 * \brief  Creates  a  PPS  from  a  Loop  subdivision  surface.   The
 * underlying mesh  of the PPS  is given by  an input OFF file,  and a
 * piecewise-linear approximation to the PPS is written out to another
 * OFF file.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 */
int main( int argc , char *argv[ ] )
{
  using std::cout ;
  using std::endl ;

  using ppsfromloop::VertexAttribute ;
  using ppsfromloop::HalfedgeAttribute ;
  using ppsfromloop::FaceAttribute ;

  //
  // typedef Surface
  //
  typedef dcel::Surface< VertexAttribute , FaceAttribute , int ,
      HalfedgeAttribute > Surface ;

  //
  // This program expects three command-line arguments.
  //
  if ( argc != 4 ) {
    cout << endl ;
    cout << "Usage: sampler-loop <arg 1> <arg 2> <arg 3>" << endl ;
    cout << endl ;
    cout << "arg1: the filename of an OFF file without extension." << endl ;
    cout << "arg2: the level-of-detail of the output triangular mesh." << endl ;
    cout << "arg3: the name of the Loop surface evaluator data table file." << endl ; 
    cout << endl ;
    cout << "For instance:" << endl ;
    cout << "sampler-loop star 3 lpdata50NT.dat" << endl ;
    cout << endl ;
    return 0 ;
  }

  /**
   * The  first  command-line  argument   is  the  name  of  the  file
   * describing the underlying mesh of the PPS, the second argument is
   * the level of  detail of the PL approximation  to the PPS computed
   * here, and the  third argument is the name  of the file containing
   * the Loop surface evaluator data table.
   */
  std::string fn1( argv[ 1 ] ) ;
  unsigned lod = unsigned( atoi( argv[ 2 ] ) ) ;
  std::string fn2( argv[ 3 ] ) ;

  //
  // Read in the underlying mesh information.
  //

  cout << "Reading input file..." << endl ;

  unsigned nv, nf ;
  double* vset ;
  unsigned* fset ;

  offlib::Reader reader( fn1 ) ;

  reader.read( nv , vset , nf , fset ) ;

  //
  // Creates the underlyng mesh and stores it in a DCEL.
  //

  cout << "Creating surface..." << endl ;

  Surface* mesh = new Surface( nv , vset , nf , fset ) ;

  //
  // Release memory.
  //

  cout << "Releasing memory..." << endl ;

  if ( vset != 0 ) {
    delete[] vset ;
  }

  if ( fset != 0 ) {
    delete[] fset ;
  }

  //
  // Creates a PPS that approximates the Loop subdivision surface.
  //

  cout << "Creating the PPS..." << endl ;

  ppsfromloop::PPSfromLOOP* pps = 
    new ppsfromloop::PPSfromLOOP( mesh , fn2 ) ;

  pps->build() ;

  //
  // Sample the Loop subdivision surface and the PPS.
  //

  cout << "Sampling the Loop subdivision surface and the PPS ..." << endl ;

  pps::PPSsampler< Surface >* sampler =  
    new pps::PPSsampler< Surface >( pps ) ;

  double* lv1 ;
  double* lv2 ;

  unsigned* lf ;

  sampler->sample( lod , nv , lv1 , lv2 , nf , lf ) ;

  //
  // Writing out the output files.
  //

  cout << "Generating the output files..." << endl ;

  std::string fn3( fn1 + std::string( "-loop-pps-" ) 
		   + std::string( argv[ 2 ] ) ) ;

  std::string fn4( fn1 + std::string( "-loop-" ) 
		   + std::string( argv[ 2 ] ) ) ;

  nv /= 3 ;
  nf /= 3 ;

  offlib::Writer writer1( fn3 ) ;
  writer1.write( nv , &lv1[ 0 ] , nf , &lf[ 0 ] ) ;

  if ( lv1 != 0 ) delete lv1 ;

  offlib::Writer writer2( fn4 ) ;
  writer2.write( nv , &lv2[ 0 ] , nf , &lf[ 0 ] ) ;

  if ( lv2 != 0 ) delete lv2 ;

  if ( lf != 0 ) delete lf ;

  //
  // Release memory.
  //

  cout << "Releasing memory..." << endl ;

  if ( sampler != 0 ) {
    delete sampler ;
  }

  if ( pps != 0 ) {
    delete pps ;
  }

  if ( mesh != 0 ) {
    delete mesh ;
  }

  return 0 ;
}
