/** 
 * \file face_attribute.h
 *  
 * \brief Definition  of the  class FaceAttribute, which  represents a
 * set of face attributes of the underlying triangle surface mesh of a
 * Parametric  Pseudo-Surface  (PPS) constructed  from  a PN  triangle
 * surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef FACE_ATTRIBUTE_H
#define FACE_ATTRIBUTE_H

#include "pntriangle.h"   // PNTriangle


/** 
* \defgroup PPSFROMPNTNameSpace Namespace ppsfrompnt.
* @{
*/

/**
* \namespace ppsfrompnt
*
* \brief  The   namespace  ppsfrompnt  contains   the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS)  that approximates the  geometry of a  given PN
* triangle surface.
*/

namespace ppsfrompnt {

  /**
   * \class FaceAttribute
   *
   * \brief This class represents a set of attributes of a face of the
   * underlying triangle surface  mesh of a PPS constructed  from a PN
   * triangle surface.
   */
  class FaceAttribute {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn FaceAttribute()
     *
     * \brief Creates an instance of this class.
     */
    FaceAttribute() : _patch( 0 )
    {}


    /**
     * \fn FaceAttribute( PNTriangle* patch )
     *
     * \brief Creates an instance of this class.
     *
     * \param patch A  pointer to the PN triangle  associated with the
     * face that owns this attribute.
     */
    FaceAttribute( PNTriangle* patch ) : _patch( patch )
    {}


    /**
     * \fn FaceAttribute( const FaceAttribute& a )
     *
     * \brief Creates an instance of this class from another instance.
     *
     * \param a A given instance of this class.
     */
    FaceAttribute( const FaceAttribute& a )
    {
      _patch = new PNTriangle( *( a.get_patch() ) ) ;
    }


    /**
     * \fn ~FaceAttribute()
     *
     * \brief Destroys an instance of this class.
     *
     */
    ~FaceAttribute() 
    {
      if ( get_patch() != 0 ) delete get_patch() ;
    }


    /**
     * \fn PNTriangle* get_patch() const
     *
     * \brief  Returns  a  pointer  to the  triangular  surface  patch
     * associated with this face.
     *
     * \return A  pointer to  the triangular surface  patch associated
     * with this face.
     */
    PNTriangle* get_patch() const
    {
      return _patch ; 
    }


    /**
     * \fn void set_patch( PNTriangle* patch )
     *
     * \brief  Assigns an  address to  the pointer  to  the triangular
     * surface patch associated with this face.
     *
     * \param  patch The address  of the  PN triangle  associated with
     * this face.
     */
    void set_patch( PNTriangle* patch )
    {
      _patch = patch ; 
    }

  private:
    // ---------------------------------------------------------------
    //
    // Private data member
    //
    // ---------------------------------------------------------------

    PNTriangle* _patch ;   ///< Pointer to the PN triangle associated with this face. 

  } ;

}

/** @} */ //end of group class.

#endif  // FACE_ATTRIBUTE_H
