/** 
 * \file halfedge_attribute.h
 *  
 * \brief Definition of  the class HalfedgeAttribute, which represents
 * a  set of halfedge  attributes of  the underlying  triangle surface
 * mesh  of a Parametric  Pseudo-Surface (PPS)  constructed from  a PN
 * triangle surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef HALFEDGE_ATTRIBUTE_H
#define HALFEDGE_ATTRIBUTE_H

#include "halfedge.h"          // dcel::Halfedge

#include "vertex_attribute.h"  // VertexAttribute
#include "face_attribute.h"    // FaceAttribute


/** 
* \defgroup PPSFROMPNTNameSpace Namespace ppsfrompnt.
* @{
*/

/**
* \namespace ppsfrompnt
*
* \brief  The   namespace  ppsfrompnt  contains   the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS)  that approximates the  geometry of a  given PN
* triangle surface.
*/

namespace ppsfrompnt {

  //
  // Forward definition
  //
  class HalfedgeAttribute ;


  /**
   * \class HalfedgeAttribute
   *
   * \brief This class  represents a set of attributes  of a half-edge
   * of the underlying triangle surface mesh of a PPS constructed from
   * a PN triangle surface.
   */
  class HalfedgeAttribute {
  public:
    // ---------------------------------------------------------------
    //
    // Type definitions
    //
    // ---------------------------------------------------------------

    /**
     * \typedef Halfedge
     *
     * \brief  Defines  Halfedge   as  an  alias  for  dcel::Halfedge<
     * VertexAttribute, FaceAttribute , int , HalfedgeAttribute >
     */    
    typedef dcel::Halfedge< 
                           VertexAttribute, 
                           FaceAttribute ,
                           int ,
                           HalfedgeAttribute
                          > 
      Halfedge ;


    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn HalfedgeAttribute()
     *
     * \brief Creates an instance of this class.
     */
    HalfedgeAttribute()
    {
      set_owner( 0 ) ;
      set_pps_id( 0 ) ;
      set_pps_id_flag( false ) ;
      set_origin_vertex_degree( 0 ) ;
      set_degree_flag( false ) ;
    }


    /**
     * \fn HalfedgeAttribute( Halfedge* h )
     *
     * \brief Creates an instance of this class.
     *
     * \param h A pointer to the halfedge that owns this attribute.
     */
    HalfedgeAttribute( Halfedge* h )
    {
      set_owner( h ) ;
      set_pps_id( 0 ) ;
      set_pps_id_flag( false ) ;
      set_origin_vertex_degree( 0 ) ;
      set_degree_flag( false ) ;
    }


    /**
     * \fn HalfedgeAttribute( HalfedgeAttribute& a )
     *
     * \brief Creates an instance of this class from another instance.
     *
     * \param a A given instance of this class.
     */
    HalfedgeAttribute( HalfedgeAttribute& a )
    {
      set_owner( a.get_owner() ) ;
      set_pps_id( a.get_pps_id() ) ;
      set_pps_id_flag( a.get_pps_id_flag() ) ;
      set_origin_vertex_degree( a.get_origin_vertex_degree() ) ;
      set_degree_flag( a.get_degree_flag() ) ;
    }


    /**
     * \fn ~HalfedgeAttribute()
     *
     * \brief Destroys an instance of this class.
     *
     */
    ~HalfedgeAttribute() 
    {
      set_owner( 0 ) ;
    }


    /**
     * \fn Halfedge* get_owner() const
     *
     * \brief  Returns  a pointer  to  the  half-edge  that owns  this
     * attribute.
     *
     * \return A pointer to the half-edge that owns this attribute.
     */
    Halfedge* get_owner() const 
    {
      return _owner ;  
    }


    /**
     * \fn void set_owner( Halfedge* h )
     *
     * \brief Assigns an address to  the pointer to the half-edge that
     * owns this attribute.
     *
     * \param h A pointer to a half-edge.
     */
    void set_owner( Halfedge* h ) 
    {
      _owner = h ;  
    }


    /**
     * \fn unsigned get_pps_id()
     *
     * \brief Returns the PPS identifier (ID) of this half-edge.  This
     * identifier is a number from 0  to N-1, where N is the degree of
     * the origin vertex of this halfedge. The number assigned to this
     * half-edge is  its position  in a counterclockwise  traversal of
     * all half-edges incident to the origin vertex of this half-edge.
     * The  first  half-edge  of   this  traversal  is  the  half-edge
     * associated with the vertex and its ID is equal to 0. The second
     * half-edge gets the ID 1, and  so on. This method is a modifier,
     * as the  ID is computed  and stored in  a data member  the first
     * time the method is invoked.
     *
     * \return The PPS identifier of this half-edge.
     */
    unsigned get_pps_id()
    {
      if ( get_pps_id_flag() ) {
	return _id ;
      }

      /**
       * To speed up  computation, compute the ID only  once and store
       * it  in  a  data  member.   Since  the  PPS  is  static,  this
       * optimization is safe.
       */
      
      set_pps_id( compute_pps_id() ) ;

      set_pps_id_flag( true ) ;

      return _id ;
    }


    /**
     * \fn void set_pps_id( unsigned id )
     *
     * \brief Assigns a value to  the data member representing the PPS
     * identifier of this half-edge.
     *
     * \param  id An  unsigned  integer representing  a half-edge  PPS
     * identifier.
     */
    void set_pps_id( unsigned id )
    {
      _id = id ;
    }


    /**
     * \fn bool get_pps_id_flag() const
     *
     * \brief Returns  the logic value  true if the PPS  identifier of
     * this  half-edge has already  been computed;  otherwise, returns
     * the logic value false.
     *
     * \return  The logic  value true  if the  PPS identifier  of this
     * half-edge  has already  been computed;  otherwise,  returns the
     * logic value false.
     */
    bool get_pps_id_flag() const
    {
      return _id_flag ;
    }


    /**
     * \fn void set_pps_id_flag( bool flag )
     *
     * \brief Assigns a  logic value (true or false)  to the flag that
     * indicates if  the PPS identifier of this  half-edge has already
     * been computed.
     *
     * \param flag A  logic value ( true or false )  to be assigned to
     * the PPS identifier flag of this half-edge.
     */
    void set_pps_id_flag( bool flag )
    {
      _id_flag = flag ;
    }


    /**
     * \fn unsigned get_origin_vertex_degree()
     *
     * \brief  Returns  the  degree  of  the  origin  vertex  of  this
     * half-edge. If  the degree has  not been computed  yet, computes
     * and  returns the  degree, which  means  that this  method is  a
     * modifier.
     *
     * \return The degree of this vertex.
     */
    unsigned get_origin_vertex_degree()
    {
      if ( get_degree_flag() ) {
	return _degree ;
      }

      /**
       * To  speed up computation,  compute the  degree only  once and
       * store it  in a  data member.  Since  the PPS is  static, this
       * optimization is safe.
       */
      
      set_origin_vertex_degree( compute_origin_vertex_degree() ) ;

      set_degree_flag( true ) ;

      return _degree ;
    }


    /**
     * \fn void set_origin_vertex_degree( unsigned degree )
     *
     * \brief  Assigns a  value to  the data  member  representing the
     * degree of the origin vertex of this half-edge.
     *
     * \param degree An unsigned integer representing a vertex degree.
     */
    void set_origin_vertex_degree( unsigned degree )
    {
      _degree = degree ;
    }


    /**
     * \fn bool get_degree_flag() const
     *
     * \brief Returns the logic value true if the degree of the origin
     * vertex of this half-edge  has already been computed; otherwise,
     * returns the logic value false.
     *
     * \return The logic value true if the degree of the origin vertex
     * of this half-edge has already been computed; otherwise, returns
     * the logic value false.
     */
    bool get_degree_flag() const
    {
      return _degree_flag ;
    }


    /**
     * \fn void set_degree_flag( bool flag )
     *
     * \brief Assigns a  logic value (true or false)  to the flag that
     * indicates if the degree of  the origin vertex of this half-edge
     * has already been computed.
     *
     * \param flag A  logic value ( true or false )  to be assigned to
     * the degree flag of this half-edge.
     */
    void set_degree_flag( bool flag )
    {
      _degree_flag = flag ;
    }


  private:
    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------
    /**
     * \fn unsigned compute_pps_id() const
     *
     * \brief  Computes and returns  the PPS  identifier (ID)  of this
     * half-edge.
     *
     * \return The PPS identifier of this half-edge.
     */
    unsigned compute_pps_id() const ;


    /**
     * \fn unsigned compute_origin_vertex_degree() const
     *
     * \brief Computes and returns the  degree of the origin vertex of
     * this half-edge.
     *
     * \return The degree of the origin vertex of this half-edge.
     */
    unsigned compute_origin_vertex_degree() const ;


    // ---------------------------------------------------------------
    //
    // Private data member
    //
    // ---------------------------------------------------------------

    Halfedge* _owner ;  ///< Pointer to the half-edge that owns this attribute..

    unsigned _id ;  ///< PPS identifier of the half-edge that owns this attribute.

    bool _id_flag ;   ///< PPS identifier flag of the half-edge that owns this attribute..

    unsigned _degree ;  ///< Degree of the origin vertex of the half-edge that owns this attribute.

    bool _degree_flag ;  ///< Degree flag of the half-edge that owns this attribute.

  } ;

}

/** @} */ //end of group class.

#endif  // HALFEDGE_ATTRIBUTE_H
