/** 
 * \file ppsfrompnt.h
 *  
 * \brief  Definition  of  the  class  PPSfromPNT  that  represents  a
 * Parametric Pseudo Surface, which approximates a PN triangle surface
 * defined on a  triangle mesh represented by a  Doubly Connected List
 * (DCEL).
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef PPSFROMPNT_H
#define PPSFROMPNT_H

#include "bezier.h"               // pps::Bezier
#include "pps.h"                  // pps::PPS

#include "surface.h"              // dcel::Surface
#include "pntriangle.h"           // PNTriangle

#include "vertex_attribute.h"     // VertexAttribute
#include "halfedge_attribute.h"   // HalfedgeAttribute
#include "face_attribute.h"       // FaceAttribute

/** 
* \defgroup PPSFROMPNTNameSpace Namespace ppsfrompnt.
* @{
*/

/**
* \namespace ppsfrompnt
*
* \brief The namespace ppsfrompnt  contains all classes that implement
* a Parametric  Pseudo-Surface (PPS)  that approximates a  PN triangle
* surface.
*
*/

namespace ppsfrompnt {

  using pps::PPS ;
  using pps::Bezier ;


  /**
   * \class PPSfromPNT
   *
   * \brief  This class  represents a  Parametric  Pseudo-Surface that
   * approximates a PN triangle surface.
   *
   */
  class PPSfromPNT : public PPS< 
                                dcel::Surface<
                                               VertexAttribute ,
                                               FaceAttribute ,
                                               int ,
                                               HalfedgeAttribute 
                                              > 
                               >
  {
  public:
    // ---------------------------------------------------------------
    //
    // Type definitions
    //
    // ---------------------------------------------------------------

    /**
     * \typedef Surface
     *
     * \brief   Defines  Surface  as   an  alias   for  dcel::Surface<
     * VertexAttribute , FaceAttribute , int , HalfedgeAttribute >
     */    
    typedef dcel::Surface< VertexAttribute , FaceAttribute , int ,
      HalfedgeAttribute > Surface ;


    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn PPSfromPNT( Surface* mesh )
     *
     * \brief Creates an instance of this PPSfromPNT class. 
     *
     * \param mesh A Pointer to a surface mesh represented by a DCEL.
     *
     */
    PPSfromPNT( Surface* mesh ) ;


    /**
     * \fn ~PPSfromPNT()
     *
     * \brief Destroys an instance of this class.
     */
    ~PPSfromPNT()
    {}


    /**
     * \fn void eval_surface( Face* face , double u , double v , double w , double& x , double& y , double& z ) const
     *
     * \brief  Computes a point  on the  PN triangle  patch associated
     * with a given face of the PPS underlying triangle mesh.
     *
     * \param face Pointer to one face of the underlying mesh.
     * \param u First barycentric coordinate of a point on the face.
     * \param v Second barycentric coordinate of a point on the face.
     * \param w Third barycentric coordinate of a point on the face.
     * \param x First Cartesian coordinate of a point on the PPS image.
     * \param y Second Cartesian coordinate of a point on the PPS image.
     * \param z Third Cartesian coordinate of a point on the PPS image.
     *
     */
    void eval_surface(
		      Face* face , 
		      double u ,
		      double v ,
		      double w ,
		      double& x ,
		      double& y ,
		      double& z
		     ) 
      const ;


  private:

    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn void build_pnt_surface()
     *
     * \brief Computes a PN triangle  surface patch for each face of a
     * given triangle mesh, and then assigns the patch to the face.
     *
     */
    void build_pnt_surface() ;

    
    /**
     * \fn void compute_vertex_normal_vector( Vertex* vertex , double& x , double& y , double& z )
     *
     * \brief Computes a unit vector at a given surface mesh vertex.
     *
     * \param vertex A pointer to a surface mesh vertex.
     * \param x A reference to the first component of the normal vector.
     * \param y A reference to the second component of the normal vector.
     * \param z A reference to the third component of the normal vector.
     */
    void compute_vertex_normal_vector(  Vertex* vertex , double& x ,
					double& y , double& z ) ;


    /**
     * \fn void compute_face_normal_vector( Face* face , double& x , double& y , double& z )
     *
     * \brief Computes the unit vector of a given surface mesh face.
     *
     * \param face A pointer to a surface mesh face.
     * \param x A reference to the first component of the normal vector.
     * \param y A reference to the second component of the normal vector.
     * \param z A reference to the third component of the normal vector.
     */
    void compute_face_normal_vector( Face* face , double& x , 
				     double& y , double& z ) ;


  public:

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn inline bool mesh_has_boundary() const
     *
     * \brief  Determines if  the underlying  mesh of  this PPS  has a
     * non-empty boundary.
     *
     * \returns  The logic  value true  if  the mesh  had a  non-empty
     * boundary and the logic value false otherwise.
     */
    inline bool mesh_has_boundary() const
    {
      return false ;
    }


    /**
     * \fn inline bool mesh_is_simplicial() const
     *
     * \brief  Determines if  the underlying  mesh  of this  PPS is  a
     * simplicial complex.
     *
     * \returns  The logic  value true  if  the mesh  is a  simplicial
     * complex and the logic value false otherwise.
     */
    inline bool mesh_is_simplicial() const
    {
      return true ;
    }


    /**
     * \fn inline unsigned int get_id( Halfedge* h ) const
     *
     * \brief Returns the identifier of a given half-edge.
     *
     * \param h Pointer to a half-edge.
     *
     * \returns The identifier of the given half-edge.
     */
    inline unsigned int get_id( Halfedge* h ) const
    {
      return h->get_attributes().get_pps_id() ;
    }


    /**
     * \fn inline Vertex* get_org( Halfedge* h ) const
     *
     * \brief Returns the  origin vertex of a given  half-edge of this
     * PPS underlying mesh.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns A pointer to the origin vertex of a given half-edge.
     */
    inline Vertex* get_org( Halfedge* h ) const
    {
      return h->get_origin() ;
    }


    /**
     * \fn inline Vertex* get_dst( Halfedge* h ) const
     *
     * \brief Returns  the destination vertex of a  given half-edge of
     * this PPS underlying mesh.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns  A  pointer  to  the  destination vertex  of  a  given
     * half-edge.
     */
    inline Vertex* get_dst( Halfedge* h )  const
    {
      return h->get_next()->get_origin() ;
    }


    /**
     * \fn inline Edge* get_edge( Halfedge* h ) const
     *
     * \brief  Returns  the  edge   a  given  half-edge  of  this  PPS
     * underlying mesh belongs to.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns A pointer to the edge a given half-edge belongs to.
     */
    inline Edge* get_edge( Halfedge* h ) const
    {
      return h->get_edge() ;
    }


    /**
     * \fn inline Face* get_face( Halfedge* h ) const
     *
     * \brief  Returns  the  face   a  given  half-edge  of  this  PPS
     * underlying mesh belongs to.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns A pointer to the face a given half-edge belongs to.
     */
    inline Face* get_face( Halfedge* h ) const
    {
      return h->get_face() ;
    }


    /**
     * \fn inline Halfedge* get_prev( Halfedge* h ) const
     *
     * \brief Returns the half-edge that precedes a given half-edge of
     * this PPS underlying  mesh in the face cycle  of half-edges that
     * contains both half-edges.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns  A pointer  to  the half-edge  that  precedes a  given
     * half-edge in their common face half-edge cycle.
     */
    inline Halfedge* get_prev( Halfedge* h ) const
    {
      return h->get_prev() ;
    }


    /**
     * \fn inline Halfedge* get_next( Halfedge* h ) const
     *
     * \brief Returns the half-edge that succeeds a given half-edge of
     * this PPS underlying  mesh in the face cycle  of half-edges that
     * contains both half-edges.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns  A pointer  to  the half-edge  that  succeeds a  given
     * half-edge in their common face half-edge cycle.
     */
    inline Halfedge* get_next( Halfedge* h ) const
    {
      return h->get_next() ;
    }
  

    /**
     * \fn inline Halfedge* get_mate( Halfedge* h ) const
     *
     * \brief  Returns the  mate  of  a given  half-edge  of this  PPS
     * underlying mesh.
     *
     * \param h Pointer to a half-edge of this PPS underlying mesh.
     *
     * \returns A pointer to the mate half-edge of a given half-edge.
     */
    inline Halfedge* get_mate( Halfedge* h ) const
    {
      return h->get_mate() ;
    }
    
    
    /**
     * \fn inline Halfedge* get_halfedge( Face* face ) const
     *
     * \brief Returns  the first half-edge of the  cycle of half-edges
     * of a given face of this PPS underlying mesh.
     *
     * \param face Pointer to a face of this PPS underlying mesh.
     *
     * \returns A pointer to the first half-edge of a given face.
     */
    inline Halfedge* get_halfedge( Face* face ) const
    {
      return face->get_halfedge() ;
    }
    
    
    /**
     * \fn inline Halfedge* get_halfedge( Vertex* vertex ) const
     *
     * \brief Returns one  half-edge with origin at a  given vertex of
     * this PPS underlying  mesh. It is assumed that  this method will
     * always return the same  half-edge (as many half-edges may share
     * the same origin vertex).
     *
     * \param vertex Pointer to a vertex of this PPS underlying mesh.
     *
     * \returns  A pointer  to one  half-edge with  origin at  a given
     * vertex.
     */
    inline Halfedge* get_halfedge( Vertex* vertex ) const
    {
      return vertex->get_halfedge() ;
    }


    /**
     * \fn inline unsigned get_degree( Vertex* vertex ) const
     *
     * \brief  Returns  the degree  of  a  given  vertex of  this  PPS
     * underlying mesh. The degree of  a vertex is the number of edges
     * incident to the vertex. 
     *
     * \param vertex Pointer to a vertex of this PPS underlying mesh.
     *
     * \returns The degree of the given vertex.
     */
    inline unsigned get_degree( Vertex* vertex ) const
    {
      return vertex->get_halfedge()->get_attributes().get_origin_vertex_degree() ;
    }

    
    /**
     * \fn inline Bezier* get_shape_function( Vertex* vertex ) const
     *
     * \brief  Returns  the shape  function  associated  with a  given
     * vertex of  this PPS  underlying mesh. The  shape function  is a
     * rectangular B&eacute;zier patch.
     *
     * \param vertex Pointer to a vertex of this PPS underlying mesh.
     *
     * \returns The shape function associated with the given vertex.
     */
    inline Bezier* get_shape_function( Vertex* vertex ) const
    {
      return vertex->get_attributes().get_patch() ;
    }
    
    
    /**
     * \fn inline void set_shape_function( Vertex* vertex, Bezier* patch )
     *
     * \brief Assigns a shape function  to a given vertex of this PPS
     * underlying mesh.
     *
     * \param vertex Pointer to a vertex of this PPS underlying mesh.
     * \param patch Pointer to a shape function.
     */
    inline void set_shape_function( Vertex* vertex, Bezier* patch )
    {
      vertex->get_attributes().set_patch( patch ) ;
    }


    /**
     * \fn inline VertexIterator vertices_begin() const
     *
     * \brief Returns a vertex iterator set to the initial vertex of a
     * vertex sequence of this PPS underlying mesh.
     *
     * \return A vertex iterator set to the initial vertex of a vertex
     * sequence of this PPS underlying mesh.
     */
    inline VertexIterator vertices_begin() const
    {
      return get_mesh()->vertices_begin() ;
    }


    /**
     * \fn inline bool is_done( const VertexIterator& iterator ) const
     *
     * \brief Returns  a logic value  true if a given  vertex iterator
     * has reached the end of a vertex sequence of this PPS underlying
     * mesh; otherwise, it returns the logic value false.
     *
     * \param iterator A vertex iterator.
     *
     * \return  A logic  value true  if  a given  vertex iterator  has
     * reached the  end of  a vertex sequence  of this  PPS underlying
     * mesh; otherwise, it returns the logic value false.
     */
    inline bool is_done( const VertexIterator& iterator ) const
    {
      return iterator == get_mesh()->vertices_end() ;
    }


    /**
     * \fn inline void move_forward( VertexIterator& iterator ) const
     *
     * \brief Makes  the iterator point  to the vertex  succeeding its
     * current  vertex in  a vertex  sequence of  this  PPS underlying
     * mesh.
     *
     * \param iterator A reference to a vertex iterator.
     */
    inline void move_forward( VertexIterator& iterator ) const
    {
      ++iterator ;
    }


    /**
     * \fn inline Vertex* get_vertex( const VertexIterator& iterator ) const
     *
     * \brief Returns  the current vertex  of a given  vertex iterator
     * for a vertex sequence of this PPS underlying mesh.
     *
     * \param iterator A reference to a vertex iterator.
     *
     * \return  A pointer  to the  current  vertex of  a given  vertex
     * iterator for a vertex sequence of this PPS underlying mesh.
     */
    inline Vertex* get_vertex( const VertexIterator& iterator ) const
    {
      return *iterator ;
    }


    /**
     * \fn inline EdgeIterator edges_begin() const
     *
     * \brief Returns an  edge iterator set to the  initial edge of an
     * edge sequence of this PPS underlying mesh.
     *
     * \return An  edge iterator  set to the  initial edge of  an edge
     * sequence of this PPS underlying mesh.
     */
    inline EdgeIterator edges_begin() const
    {
      return get_mesh()->edges_begin() ;
    }


    /**
     * \fn inline bool is_done( const EdgeIterator& iterator ) const
     *
     * \brief Returns A logic value  true if a given edge iterator has
     * reached  the end  of an  edge sequence  of this  PPS underlying
     * mesh; otherwise, it returns the logic value false.
     *
     * \param iterator An edge iterator.
     *
     * \return A logic value true if a given edge iterator has reached
     * the  end of  an  edge  sequence of  this  PPS underlying  mesh;
     * otherwise, it returns the logic value false.
     */
    inline bool is_done( const EdgeIterator& iterator ) const
    {
      return iterator == get_mesh()->edges_end() ;
    }


    /**
     * \fn inline void move_forward( EdgeIterator& iterator ) const
     *
     * \brief  Makes the  iterator point  to the  edge  succeeding its
     * current edge in an edge sequence of this PPS underlying mesh.
     *
     * \param iterator A reference to an edge iterator.
     */
    inline void move_forward( EdgeIterator& iterator ) const
    {
      ++iterator ;
    }


    /**
     * \fn inline Edge* get_edge( const EdgeIterator& iterator ) const
     *
     * \brief Returns the current edge of a given edge iterator for an
     * edge sequence of this PPS underlying mesh.
     *
     * \param iterator A reference to an edge iterator.
     *
     * \return A pointer to the  current edge of a given edge iterator
     * for an edge sequence of this PPS underlying mesh.
     */
    inline Edge* get_edge( const EdgeIterator& iterator ) const
    {
      return *iterator ;
    }


    /**
     * \fn inline FaceIterator faces_begin() const
     *
     * \brief Returns  a face  iterator set to  the initial face  of a
     * face sequence of this PPS underlying mesh.
     *
     * \return  A face  iterator set  to the  initial face  of  a face
     * sequence of this PPS underlying mesh.
     */
    inline FaceIterator faces_begin() const
    {
      return get_mesh()->faces_begin() ;
    }


    /**
     * \fn inline bool is_done( const FaceIterator& iterator ) const
     *
     * \brief Returns A logic value  true if a given face iterator has
     * reached the end of a face sequence of this PPS underlying mesh;
     * otherwise, it returns the logic value false.
     *
     * \param iterator A face iterator.
     *
     * \return A logic value true if a given face iterator has reached
     * the  end  of a  face  sequence  of  this PPS  underlying  mesh;
     * otherwise, it returns the logic value false.
     */
    inline bool is_done( const FaceIterator& iterator ) const
    {
      return iterator == get_mesh()->faces_end() ;
    }


    /**
     * \fn inline void move_forward( FaceIterator& iterator ) const
     *
     * \brief  Makes the  iterator point  to the  face  succeeding its
     * current face in a face sequence of this PPS underlying mesh.
     *
     * \param iterator A reference to a face iterator.
     */
    inline void move_forward( FaceIterator& iterator ) const
    {
      ++iterator ;
    }


    /**
     * \fn inline Face* get_face( const FaceIterator& iterator ) const
     *
     * \brief Returns the current face  of a given face iterator for a
     * face sequence of this PPS underlying mesh.
     *
     * \param iterator A reference to a face iterator.
     *
     * \return A pointer to the  current face of a given face iterator
     * for a face sequence of this PPS underlying mesh.
     */
    inline Face* get_face( const FaceIterator& iterator ) const
    {
      return *iterator ;
    }

  } ;


}

/** @} */ //end of group class.

#endif   // PPSFROMPNT_H
